/*
 * RT-Linux mutex implementation
 *
 * Written by Michael Barabanov
 * Copyright (C) VJY Associates LLC, 1999
 * Released under the terms of the GPL
 *
 */

#ifndef __RTL_MUTEX__
#define __RTL_MUTEX__

#ifdef __KERNEL__

#include <rtl_conf.h>
#ifdef CONFIG_RTL_MUTEX_SUPPORT

#include <rtl_sched.h>
#include <rtl_waitqueue.h>

enum { PTHREAD_MUTEX_NORMAL, PTHREAD_MUTEX_RECURSIVE, PTHREAD_MUTEX_ERRORCHECK,
	      PTHREAD_MUTEX_DEFAULT = PTHREAD_MUTEX_NORMAL };

typedef struct {
	int type;
	int pshared;
} pthread_mutexattr_t;

enum { PTHREAD_PROCESS_SHARED, PTHREAD_PROCESS_PRIVATE };

extern inline int pthread_mutexattr_getpshared(const pthread_mutexattr_t *attr,
		    int *pshared)
{
	*pshared = attr->pshared;
	return 0;
}

extern inline int pthread_mutexattr_setpshared(pthread_mutexattr_t *attr,
		    int pshared)
{
	if (pshared != PTHREAD_PROCESS_SHARED
			&& pshared != PTHREAD_PROCESS_PRIVATE) {
		return EINVAL;
	}
	attr->pshared = pshared;
	return 0;
}


extern inline int pthread_mutexattr_init(pthread_mutexattr_t *attr)
{
	attr->type = PTHREAD_MUTEX_DEFAULT;
	attr->pshared = PTHREAD_PROCESS_PRIVATE;
	return 0;
}

extern inline int pthread_mutexattr_destroy(pthread_mutexattr_t *attr)
{
	return 0;
}

extern int pthread_mutexattr_settype(pthread_mutexattr_t *attr, int type);

extern inline int pthread_mutexattr_gettype(const pthread_mutexattr_t *attr, int *type)
{
	*type = attr->type;
	return 0;
}


typedef struct {
	int count;
	spinlock_t lock;
	int type;
	RTL_WQ wq;
} pthread_mutex_t;



#define PTHREAD_MUTEX_INITIALIZER { 0, SPIN_LOCK_UNLOCKED, PTHREAD_MUTEX_DEFAULT, RTL_WQ_INITIALIZER }

extern int pthread_mutex_init(pthread_mutex_t *mutex,
		    const pthread_mutexattr_t *attr);
extern int pthread_mutex_destroy(pthread_mutex_t *mutex);

extern int pthread_mutex_lock(pthread_mutex_t *mutex);
extern int pthread_mutex_trylock(pthread_mutex_t *mutex);
extern int pthread_mutex_unlock(pthread_mutex_t *mutex);

#endif
#endif


/* Common routines for implementing IPC */

extern inline void rtl_thread_mark_self_suspended(void)
{
	pthread_self()->state = RTL_THREAD_SUSPENDED;
}

extern inline void rtl_thread_pend_wakeup(pthread_t thread)
{
	set_bit (RTL_SIGNAL_WAKEUP, &thread->pending_signals);
}

extern inline void rtl_reschedule2(int cpu)
{
#ifdef __SMP__
	if (cpu != rtl_getcpuid())
		rtl_reschedule(cpu);
	else
#endif
		rtl_schedule();
}

#endif
