/*
 * RT-Linux mutex implementation
 *
 * Written by Michael Barabanov
 * Copyright (C) VJY Associates LLC, 1999
 * Released under the terms of the GPL
 *
 */

#include <rtl_mutex.h>
#ifdef CONFIG_RTL_MUTEX_SUPPORT

#include <rtl_sync.h>

int pthread_mutexattr_settype(pthread_mutexattr_t *attr, int type)
{
	if (attr->type != PTHREAD_MUTEX_DEFAULT) {
		/* TODO: implement various mutex types */
		return EINVAL;
	}
	attr->type = type;
	return 0;
}

int pthread_mutex_init(pthread_mutex_t *mutex,
		    const pthread_mutexattr_t *attr)
{
	rtl_wq_init (&mutex->wq);
	spin_lock_init (&mutex->lock);
	mutex->count = 0;
	if (attr) {
		pthread_mutexattr_gettype(attr, &mutex->type);
	} else {
		mutex->type = PTHREAD_MUTEX_DEFAULT;
	}
	return 0;
}

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	return 0;
}


/* TODO: implement various mutex types */
int __pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	if (test_and_set_bit(0, &mutex->count)) {
		return EBUSY;
	}
	return 0;
}

int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	rtl_irqstate_t flags;
	int ret;
	rtl_spin_lock_irqsave (&mutex->lock, flags);

	ret = __pthread_mutex_trylock(mutex);

	rtl_spin_unlock_irqrestore (&mutex->lock, flags);
	return ret;
}

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
	rtl_irqstate_t flags;
	int ret;
	rtl_spin_lock_irqsave (&mutex->lock, flags);

	ret = __pthread_mutex_trylock(mutex);
	if (ret != EBUSY) {

		rtl_spin_unlock_irqrestore (&mutex->lock, flags);
		return ret;
	}

	if (rtl_wq_insert(&mutex->wq, pthread_self())) {
		rtl_spin_unlock_irqrestore (&mutex->lock, flags);
		return EAGAIN;
	}
	rtl_thread_mark_self_suspended();

	rtl_spin_unlock_irqrestore (&mutex->lock, flags);
	rtl_schedule();
	return 0;
}


int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	rtl_irqstate_t flags;
	pthread_t thread;
	rtl_spin_lock_irqsave (&mutex->lock, flags);

	thread = rtl_wq_remove (&mutex-> wq);
	if (thread) {
		rtl_thread_pend_wakeup(thread);
	} else {
		clear_bit (0, &mutex->count);
		rtl_spin_unlock_irqrestore (&mutex->lock, flags);
		return 0;
	}
	rtl_spin_unlock_irqrestore (&mutex->lock, flags);
	rtl_reschedule2 (thread->cpu);
	return 0;
}


#endif
